import base64
import bz2
import json
from functools import lru_cache
from typing import Dict

from async_lru import alru_cache
from ldap3.utils.dn import to_dn

from ucsschool.apis.utils import LDAPCredentials, LDAPSettings
from udm_rest_client import UDM, NoObject


class IdBrokerPseudonymisationError(Exception):
    ...


class IdBrokerPseudonymisationNoMapping(IdBrokerPseudonymisationError):
    ...


class IdBrokerPseudonymisationNoSecret(IdBrokerPseudonymisationError):
    ...


class IdBrokerPseudonymisationDataError(IdBrokerPseudonymisationError):
    ...


@lru_cache(maxsize=1)
def ldap_settings() -> LDAPSettings:
    """Cached LDAP settings object"""
    return LDAPSettings()


@lru_cache(maxsize=1)
def ldap_credentials() -> LDAPCredentials:
    """Cached LDAP credentials object"""
    return LDAPCredentials(ldap_settings())


@lru_cache(maxsize=1)
def udm_kwargs() -> Dict[str, str]:
    """Cached UDM REST API client credentials using the Docker apps host DN."""
    _ldap_settings = ldap_settings()
    _ldap_credentials = LDAPCredentials(_ldap_settings)
    username = to_dn(_ldap_credentials.host_dn, decompose=True)[0][1]
    username = f"{username}$"
    password = _ldap_credentials.machine_password
    return {
        "username": username,
        "password": password,
        "url": f"https://{_ldap_settings.host_fqdn}/univention/udm/",
        "ssl_ca_cert": "/etc/ssl/certs/ca-certificates.crt",
    }


async def udm_ctx() -> UDM:
    """
    Can be used in a FastAPI resource as dependency injection like this:

    async def ...(..., udm: UDM = Depends(udm_ctx)) -> ...:
        async for udm_obj in udm.get("users/user").search(...):
            ...
    """
    async with UDM(**udm_kwargs()) as udm:
        yield udm


def decode_settings_data(value: bytes) -> bytes:
    return bz2.decompress(base64.b64decode(value))


async def get_settings_data_content(dn: str) -> Dict[str, str]:
    """
    Get `data` attribute of a `settings/data` object.

    :raises IdBrokerPseudonymisationDataError: when `settings/data` UDM object does not exist or content
        is not valid JSON
    """
    async with UDM(**udm_kwargs()) as udm:
        mod = udm.get("settings/data")
        try:
            obj = await mod.get(dn)
        except NoObject as exc:
            raise IdBrokerPseudonymisationDataError(f"Loading {dn!r}: {exc!s}") from exc
        try:
            raw_data = decode_settings_data(obj.props.data)
            data = json.loads(raw_data)
            if not isinstance(data, dict):
                raise ValueError(f"Expected a dict, found: {type(data)!r}")
        except ValueError as exc:
            raise IdBrokerPseudonymisationDataError(f"Invalid data in {dn!r}: {exc!s}") from exc
        return data


@alru_cache(maxsize=30)
async def _cached_get_service_provider_attribute(s_p: str) -> str:
    """Attention: Cache will also contain exceptions for missing service providers."""
    ldap_base = ldap_settings().ldap_base
    dn = f"cn=IDBrokerServiceProviderMappings,cn=IDBrokerSettings,cn=univention,{ldap_base}"
    data = await get_settings_data_content(dn)
    try:
        return data[s_p]
    except KeyError as exc:
        raise IdBrokerPseudonymisationNoMapping(
            f"No attribute mapping for SP {s_p!r} found. Known mappings: " f"{sorted(data.keys())!r}"
        ) from exc


async def get_service_provider_attribute(service_provider: str) -> str:
    """
    Get LDAP/UDM attribute name for pseudonyms for service provider.

    Only successful lookups will be cached, so the admin can later add a service provider and the API
    server will not have to be restarted.

    :raises IdBrokerPseudonymisationDataError: when `settings/data` UDM object does not exist or content
        is not valid JSON
    :raises IdBrokerPseudonymisationNoMapping: when `service_provider` does not exist in `settings/data`
        UDM object
    """
    try:
        return await _cached_get_service_provider_attribute(service_provider)
    except IdBrokerPseudonymisationNoMapping:
        # Remove exception from cache. We only cache successful lookups.
        _cached_get_service_provider_attribute.invalidate(service_provider)
        raise


@alru_cache(maxsize=30)
async def _cached_get_service_provider_secret(s_p: str) -> str:
    """Attention: Cache will also contain exceptions for missing service providers."""
    ldap_base = ldap_settings().ldap_base
    dn = f"cn=IDBrokerServiceProviderSecrets,cn=IDBrokerSettings,cn=univention,{ldap_base}"
    data = await get_settings_data_content(dn)
    try:
        return data[s_p]
    except KeyError as exc:
        raise IdBrokerPseudonymisationNoSecret(
            f"No secret for SP {s_p!r} found. Known SPs: {sorted(data.keys())!r}"
        ) from exc


async def get_service_provider_secret(service_provider: str) -> str:
    """
    Get secret for service provider.

    Only successful lookups will be cached, so the admin can later add a service provider and the API
    server will not have to be restarted.

    :raises IdBrokerPseudonymisationDataError: when `settings/data` UDM object does not exist or content
        is not valid JSON
    :raises IdBrokerPseudonymisationNoSecret: when `service_provider` does not exist in `settings/data`
        UDM object
    """
    try:
        return await _cached_get_service_provider_secret(service_provider)
    except IdBrokerPseudonymisationNoSecret:
        # Remove exception from cache. We only cache successful lookups.
        _cached_get_service_provider_secret.invalidate(service_provider)
        raise
